#include <OS.h>
#include <Roster.h>
#include <Message.h>
#include <stdlib.h>
#include <stdio.h>
#include "Scripting.h"
#undef max
#undef min
extern "C" {
#define __STDC__
#include "error.h"
//#undef __cplusplus
#include "_scm.h"
//#define __cplusplus
};
#include "script.h"

const false = 0;
const true = 1;

	static
	bool GetApp(char *name, BMessenger &outMsgr)
	{
		BList teamList;
		be_roster->GetAppList(&teamList);
		bool gotIt = false;
		for (int ix=teamList.CountItems()-1; ix>=0; ix--) {
			team_id team = (team_id)teamList.ItemAt(ix);
			thread_info info;
			get_nth_thread_info(team, 0, &info);
			if (!strcmp(info.name, name)) {
				team_info tm_info;
				get_team_info(team, &tm_info);
				BMessenger ret(0, team);
				outMsgr = ret;
				gotIt = true;
				break;
			}
		}
		if (!gotIt) {
			fprintf(stderr, "Can't find app: %s\n", name);
			return false;
		}
		return true;
	}


	static SCM
	get_property(SCM list, PropertyItem &prop)
	{
		SCM scm = SCM_CAR(list);
		SCM_ASSERT(SCM_NIMP(scm) && SCM_ROSTRINGP(scm), scm,
			SCM_ARG3, s_tell);
		char *str = SCM_CHARS(scm);
		
		strncpy(prop.property, str, kPropNameLength-1);
		prop.form = formDirect;
		
		list = SCM_CDR(list);

		if (SCM_NIMP(list) && SCM_CONSP(list)) {
			scm = SCM_CAR(list);
			
			if (SCM_NUMBERP(scm)) {
				long ix = scm_num2long(scm, (char *)SCM_ARG3, s_tell);
				
				prop.data.index = ix;
				if (ix < 0) {
					prop.form = formReverseIndex;
				}
				else {
					prop.form = formIndex;
				}
				list = SCM_CDR(list);
			}
			else if (SCM_ROSTRINGP(scm)) {
				char *str = SCM_CHARS(scm);
				strncpy(prop.data.name, str, kPropNameLength-1);
				prop.form = formName;
				list = SCM_CDR(list);
			}
		}
		return list;
	}	
				

SCM_PROC (s_tell, "tell", 2, 1, 0, scm_script_tell);

SCM 
scm_script_tell (
	SCM app,
	SCM verb,
	SCM props)
{
	char *appstr;
	long verbcode;

	SCM_ASSERT (SCM_NIMP (app) && SCM_ROSTRINGP (app), app,
	      SCM_ARG1, s_tell);
	appstr = SCM_CHARS (app);

    SCM_ASSERT (SCM_NUMBERP (verb), verb, SCM_ARG2, s_tell);
	verbcode = scm_num2long(verb, (char *)SCM_ARG2, s_tell); //SCM_INUM (verb);	


	BMessage *reply = 0;
	BMessenger msgr;
	
	// get application
	if (! GetApp(appstr, msgr)) {
		return SCM_BOOL_F;
	}

	BMessage *msg = new BMessage(verbcode);

	SCM prop, propname, propvalue;
	char *propnamestr = 0;
	
	for (; SCM_NIMP(props); props = SCM_CDR(props)) {
		SCM_ASSERT(SCM_CONSP(props), props, SCM_ARG3, s_tell);
		prop = SCM_CAR(props);
		SCM_ASSERT(SCM_NIMP(prop) && SCM_CONSP(prop), props, SCM_ARG3, s_tell);

		propname = SCM_CAR(prop);
		SCM_ASSERT(SCM_NIMP(propname) && SCM_ROSTRINGP(propname), propname,
			SCM_ARG3, s_tell);
		propnamestr = SCM_CHARS(propname);
		
		propvalue = SCM_CDR(prop);
	
		while (SCM_NIMP(propvalue)) {	
			// is it a list?
			if (SCM_CONSP (propvalue)) {
				// of list
				SCM list = SCM_CAR (propvalue);
				if (SCM_CONSP (list)) {
					// assume property
					PropertyItem property;
					propvalue = get_property(list, property);
					msg->AddData(propnamestr, PROPERTY_TYPE, &property, sizeof(property));
				}
				else {
					propvalue = list; // check this
				}
			}
			else {
				// no list but
				// what type is it?
				if (SCM_INUMP (propvalue)) {
					// number
					long num = SCM_INUM (propvalue);
					msg->AddLong (propnamestr, num);
				}
				else if (SCM_NIMP (propvalue) && SCM_ROSTRINGP (propvalue)) {
					// string
					char *str = SCM_CHARS (propvalue);
					msg->AddString(propnamestr, str);
				}
				break;
			}
		}
	}	

	if (msgr.SendMessage(msg, &reply)) {
		return SCM_BOOL_F;
	}

	{
		unsigned long type;
		long count;
		
//		if (reply->GetInfo(kDefaultErrorName, &type, &count)) {
//		}	

		if (!msg->GetInfo(kDefaultDataName, &type, &count)) {
			fprintf(stderr, "(empty data)\n");	
		}
		else {
			for (int ix=0; ix<count; ix++) {
				fprintf(stderr, "%5d: ", ix);
				switch (type) {
				case B_LONG_TYPE:
					fprintf(stderr, "%d\n", msg->FindLong(kDefaultDataName, ix));
					break;
				case B_FLOAT_TYPE:
					fprintf(stderr, "%f\n", msg->FindFloat(kDefaultDataName, ix));
					break;
				case B_RECT_TYPE:
					BRect r = msg->FindRect(kDefaultDataName, ix);
					fprintf(stderr, "%f,%f,%f,%f\n", r.left, r.top, r.right, r.bottom);
					break;
				case B_STRING_TYPE:
					fprintf(stderr, "'%s'\n", msg->FindString(kDefaultDataName, ix));
					break;
				case B_ASCII_TYPE:
					{
						long size = 0;
						void *ptr = msg->FindData(kDefaultDataName, type, ix, &size);
						fprintf(stderr, "'%.*s'\n", size, ptr);
					}
					break;
				case B_BOOL_TYPE:
					fprintf(stderr, "%s\n", msg->FindBool(kDefaultDataName)?"true":"false");
					break;
				default:
					fprintf(stderr, "<unknown type>\n");
					break;
				}
			}
		}
	}

//	fprintf(stderr, "'%s' '%s'\n", appstr, verbstr);
	
	delete reply;

	return SCM_BOOL_T;
}

void
scm_init_script ()
{
#include "script.x"
}

